/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose.mdm;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

import de.bsvrz.ibv.simbaustelle.prognose.BerechnungsZeitRaumSchnittstelle;
import de.bsvrz.ibv.simbaustelle.verwaltung.SimulationsAuftrag;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Bestimmung des Berechungszeitraums für eine Mdm-Baustelle.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class MDMBerechnungsZeitRaum implements BerechnungsZeitRaumSchnittstelle {

	private static final Debug LOGGER = Debug.getLogger();

	private long startZeitPunkt;
	private long endeZeitPunkt;
	private List<Long> intervalle = new ArrayList<>();

	private Long schrittWeiteInMsec;

	public MDMBerechnungsZeitRaum(SimulationsAuftrag auftrag) throws UndefiniertException {

		final Calendar cal = Calendar.getInstance();

		// Anfangszeitpunkt ermitteln
		cal.setTimeInMillis(auftrag.getStartZeit());
		cal.set(Calendar.HOUR_OF_DAY, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);

		long horizontZeit = cal.getTimeInMillis();
		LOGGER.finer("Auftrags-Startzeit: " + DateFormat.getDateTimeInstance().format(cal.getTime()));

		cal.setTimeInMillis(auftrag.getBaustelle().getDaten().getStart());
		cal.set(Calendar.HOUR_OF_DAY, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);

		long bstZeit = cal.getTimeInMillis();
		LOGGER.finer("Baustellen-Startzeit: " + DateFormat.getDateTimeInstance().format(cal.getTime()));

		startZeitPunkt = Math.max(horizontZeit, bstZeit);
		LOGGER.finer("Prognosestart: " + DateFormat.getDateTimeInstance().format(new Date(startZeitPunkt)));

		// Endzeitpunkt ermitteln
		cal.setTimeInMillis(auftrag.getEndZeit());
		cal.set(Calendar.HOUR_OF_DAY, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);

		horizontZeit = cal.getTimeInMillis();
		LOGGER.finer("Auftrags-Endzeit: " + DateFormat.getDateTimeInstance().format(cal.getTime()));

		cal.setTimeInMillis(auftrag.getBaustelle().getDaten().getEnde());
		cal.set(Calendar.HOUR_OF_DAY, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);

		bstZeit = cal.getTimeInMillis();
		LOGGER.finer("Baustellen-Endzeit: " + DateFormat.getDateTimeInstance().format(cal.getTime()));

		endeZeitPunkt = Math.min(horizontZeit, bstZeit) + TimeUnit.DAYS.toMillis(1);
		LOGGER.finer("Prognoseende: " + DateFormat.getDateTimeInstance().format(new Date(endeZeitPunkt)));

		schrittWeiteInMsec = auftrag.getSchrittWeiteInMsec();
		if ((schrittWeiteInMsec == null) || (schrittWeiteInMsec <= 0)) {
			schrittWeiteInMsec = TimeUnit.HOURS.toMillis(1);
		}

		if (endeZeitPunkt > startZeitPunkt) {
			long schrittZeitPunkt = startZeitPunkt + schrittWeiteInMsec;
			while (schrittZeitPunkt <= endeZeitPunkt) {
				intervalle.add(schrittZeitPunkt);
				schrittZeitPunkt += schrittWeiteInMsec;
			}
		}
	}

	@Override
	public long getEndeZeitpunkt() {
		return endeZeitPunkt;
	}

	@Override
	public List<Long> getIntervalle() {
		return intervalle;
	}

	@Override
	public long getStartZeitpunkt() {
		return startZeitPunkt;
	}

	@Override
	public boolean isValid() {
		return prognoseMoeglich();
	}

	@Override
	public boolean prognoseMoeglich() {
		return endeZeitPunkt > startZeitPunkt && !intervalle.isEmpty();
	}

	@Override
	public String toString() {
		final StringBuffer buffer = new StringBuffer();
		buffer.append('(');
		buffer.append(DateFormat.getDateTimeInstance().format(new Date(startZeitPunkt)));
		buffer.append('-');
		buffer.append(DateFormat.getDateTimeInstance().format(new Date(endeZeitPunkt)));
		buffer.append(", Schrittweite: ");
		buffer.append(schrittWeiteInMsec);
		buffer.append(" ms)");
		return buffer.toString();
	}

	@Override
	public long getSchrittweiteInMsec() {
		if (schrittWeiteInMsec == null) {
			return 0;
		}
		return schrittWeiteInMsec;
	}
}
