/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.verwaltung;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssKonstanten;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * das Modul zur Verwaltung der Baustellen eines VerkehrsmodellNetz.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BaustellenVerwaltung implements MutableSetChangeListener {

	/**
	 * der zyklische Task zur Berechnung der Prognosedaten für alle Baustellen.
	 *
	 * @author peuker
	 */
	public class BaustellenBerechner extends TimerTask {

		/**
		 * das Verwaltungsmodul für das die Berechnungen ausgeführt werden sollen.
		 */
		private final BaustellenVerwaltung verwaltung;

		/**
		 * erzeugt eine Instanz der Berechnungstask.
		 *
		 * @param verwaltung
		 *            die zu bedienende Baustellenverwaltung
		 */
		public BaustellenBerechner(final BaustellenVerwaltung verwaltung) {
			this.verwaltung = verwaltung;
		}

		/**
		 * DIe Funktion führt eine Berechnung aller Baustellen für den aktuellen
		 * Zeitpunkt aus.
		 *
		 * @see java.util.TimerTask#run()
		 */
		@Override
		public final void run() {
			verwaltung.berechneBaustellen(System.currentTimeMillis());
		}

	}

	/**
	 * der Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * das Baustellensimulationsprogramm das das Verwaltungsmodul angelegt hat.
	 */
	// private final BaustellenSimulation simulation;

	/**
	 * die Liste der Straßensegmente, die das VerkehrsmodellNetz bilden.
	 */
	private final Set<StrassenSegment> segmentListe = new HashSet<>();

	/** die Menge der Baustellen, die verwaltet werden. */
	private final Map<Long, BssBaustelle> baustellen = new HashMap<>();

	/** die Menge der MDM-Baustellen, die verwaltet werden. */
	private final Map<Long, MdmBaustelle> mdmBaustellen = new HashMap<>();

	private Timer automatischeBerechnungTimer;

	private MutableSet baustellenMenge;

	private MutableSet mdmBaustellenMenge;

	private VerkehrModellNetz netz;

	private ConfigurationObject mdmVerwaltung;

	private SimulationsAuftragsVerwalter auftragsVerwaltung = new SimulationsAuftragsVerwalter();

	private IBssParameterObserver parameterObserver;

	/**
	 * erzeugt eine INstanz des Verwaltungsmoduls.
	 *
	 * @param simulation
	 *            die aufrufende Baustellensimulation
	 * @param netz
	 *            das zu bedienende Netz
	 * @param mdmVerwaltung
	 *            das Objekt mit dem die MDM-Baustellen verwaltet werden
	 */
	public BaustellenVerwaltung(final BaustellenSimulation simulation, final VerkehrModellNetz netz,
			final ConfigurationObject mdmVerwaltung) {
		parameterObserver = simulation.getParameter();
		this.netz = netz;
		this.mdmVerwaltung = mdmVerwaltung;
	}

	/** initialisiert die Baustellenverwaltung. */
	public void init() {

		auftragsVerwaltung.start();

		if (netz != null) {
			segmentListe.addAll(netz.getNetzSegmentListe());

			baustellenMenge = ((ConfigurationObject) netz.getSystemObject())
					.getMutableSet(BssKonstanten.ATT_NAME_FELD_NETZ_BAUSTELLEN);
			baustellenMenge.addChangeListener(this);

			for (final SystemObject bst : baustellenMenge.getElements()) {
				synchronized (baustellen) {
					baustellen.put(bst.getId(), new BssBaustelle(segmentListe, bst));
				}
			}
			startBackgroundSimulation();
		}

		if (mdmVerwaltung != null) {
			mdmBaustellenMenge = mdmVerwaltung.getMutableSet("MengeMdmBaustellen");
			mdmBaustellenMenge.addChangeListener(this);

			for (final SystemObject bst : mdmBaustellenMenge.getElements()) {
				synchronized (mdmBaustellen) {
					mdmBaustellen.put(bst.getId(), new MdmBaustelle(bst));
				}
			}
		}

		LOGGER.fine("Verwaltungsmodul initialisiert");
	}

	/** startet die automatische Simulation der Baustellen. */
	private void startBackgroundSimulation() {
		if (automatischeBerechnungTimer != null) {
			return;
		}

		automatischeBerechnungTimer = new Timer();
		automatischeBerechnungTimer.scheduleAtFixedRate(new BaustellenBerechner(this), Constants.MILLIS_PER_MINUTE,
				Constants.MILLIS_PER_HOUR);
		LOGGER.fine("Automatische Baustellensimulation gestartet");
	}

	/**
	 * führt eine zyklische Berechnung aller Baustellen aus und aktualisiert deren
	 * Berechungszeitpunkt.
	 *
	 * @param zeit
	 *            der Berechnungszeitpunkt
	 */
	public final void berechneBaustellen(final long zeit) {
		LOGGER.finest("Berechne Baustellenliste");
		synchronized (baustellen) {
			for (final BssBaustelle baustelle : baustellen.values()) {
				if ((baustelle.getLetzteZyklischeBerechnung()
						+ parameterObserver.getAktualisierungsIntervall()) <= zeit) {
					BaustellenSimulation.getInstance()
							.addAuftrag(new ZyklischerAuftrag(baustelle, "System", "Automatische Simulation", null));
					baustelle.aktualisiereLetzteZyklischeBerechnung();
				}
			}
		}
	}

	/**
	 * liefert eine Kopie der Liste der verwalteten Baustellen.
	 *
	 * @return die Liste
	 */
	public final Collection<BssBaustelle> getBaustellenListe() {
		return new ArrayList<>(baustellen.values());
	}

	/**
	 * Die Funktion empfängt Änderungen der Liste der Baustellen des bedienten
	 * Netzes und erzeugt die erforderlichen Baustellenobjekte bzw. entfernt diese.
	 *
	 * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet,
	 *      de.bsvrz.dav.daf.main.config.SystemObject[],
	 *      de.bsvrz.dav.daf.main.config.SystemObject[])
	 */
	@Override
	public final void update(final MutableSet set, final SystemObject[] addedObjects,
			final SystemObject[] removedObjects) {

		LOGGER.finer("Baustellenliste wurde geändert!");

		if (set.equals(baustellenMenge)) {
			synchronized (baustellen) {
				if (removedObjects != null) {
					for (final SystemObject obj : removedObjects) {
						LOGGER.finer("Baustelle " + obj + " entfernen");
						final BssBaustelle bst = baustellen.remove(obj.getId());
						bst.disconnect();
						LOGGER.finer("Baustellenobjekt für " + obj + " entfernt");
					}
				}
				if (addedObjects != null) {
					for (final SystemObject obj : addedObjects) {
						LOGGER.finer("Baustelle " + obj + " hinzufügen");
						baustellen.put(obj.getId(), new BssBaustelle(segmentListe, obj));
						LOGGER.finer("Baustellenobjekt für " + obj + " angelegt");
					}
				}
			}
		} else if (set.equals(mdmBaustellenMenge)) {
			synchronized (mdmBaustellen) {
				if (removedObjects != null) {
					for (final SystemObject obj : removedObjects) {
						LOGGER.finer("MDM-Baustelle " + obj + " entfernen");
						final MdmBaustelle bst = mdmBaustellen.remove(obj.getId());
						bst.disconnect();
						LOGGER.finer("MDM-Baustellenobjekt für " + obj + " entfernt");
					}
				}
				if (addedObjects != null) {
					for (final SystemObject obj : addedObjects) {
						LOGGER.finer("MDM-Baustelle " + obj + " hinzufügen");
						mdmBaustellen.put(obj.getId(), new MdmBaustelle(obj));
						LOGGER.finer("MDM-Baustellenobjekt für " + obj + " angelegt");
					}
				}
			}
		}
	}

	/**
	 * fügt der Baustellenverwaltung einen neuen Auftrag hinzu.
	 * 
	 * @param auftrag
	 *            der neue Auftrag
	 */
	public void addAuftrag(SimulationsAuftrag auftrag) {
		auftragsVerwaltung.addAuftrag(auftrag);
	}
}
